import 'package:flutter/material.dart';
import 'utils/holiday.dart';
import 'utils/lunar.dart';
import 'utils/solar.dart';

typedef FxSelectedDateTime = void Function(DateTime selDateTime);
/// 节假日
mixin FxHolidayDelegate {
  Holiday holiday;
  /// -1正常上班，1调班，0休假
  int get workType {
    return holiday == null ? -1 : (holiday.work ? 1 : 0);
  }
  // 节假日显示
  String get holidayName;
}
/// 农历
mixin FxLunarDelegate {
  Lunar lunar;
  /// 日历显示内容
  String get lunarName;
}

mixin FxSolarDelegate {
  Solar solar;
  String get solarName;
}

/// 每天显示的数据模型
class FxItemModel {
  /// 外部用户配置参数
  FxItemModel({
    this.normalBackColor = Colors.white,
    this.selectedBackColor = Colors.green,
    this.normalTitleColor = Colors.black,
    this.selectedTitleColor = Colors.white,
    this.otherMonthTitleColor = Colors.black54,
    this.isShowLastNextMonth = true,
    this.isShowHoliday = false,
    this.isShowLunarCalendar = false,
    this.isShowSolarCalendar = false
  });
  /// 工厂方法 - 显示节假日
  factory FxItemModel.more({
    Color normalBackColor,
    Color selectedBackColor,
    Color normalTitleColor,
    Color selectedTitleColor,
    Color otherMonthTitleColor,
    bool isShowLastNextMonth,
    bool isShowSolarCalendar,
    Holiday holiday,
    Lunar lunar,
    Solar solar
  }) = FxItemDetailModel;

  /// 正常显示的背景颜色
  final Color normalBackColor;
  /// 选中显示的背景颜色
  final Color selectedBackColor;
  /// 当月正常显示的文字颜色
  final Color normalTitleColor;
  /// 当月选中显示的文字颜色
  final Color selectedTitleColor;
  /// 上月/下月显示的文字颜色
  final Color otherMonthTitleColor;
  /// 是否显示上月/下月的日期
  final bool isShowLastNextMonth;
  /// 是否显示休假
  final bool isShowHoliday;
  /// 是否显示农历
  final bool isShowLunarCalendar;
  /// 是否显示24节气
  final bool isShowSolarCalendar;

  /// 当前日期是否是上月/下月的日期
  bool isLastNextMonth;

  int year;
  int month;
  int day;

  int get fxDateCount {
    return year * 1000 + month * 100 + day;
  }
}

class FxItemDetailModel extends FxItemModel with FxHolidayDelegate, FxLunarDelegate, FxSolarDelegate {

  FxItemDetailModel({
    Color normalBackColor,
    Color selectedBackColor,
    Color normalTitleColor,
    Color selectedTitleColor,
    Color otherMonthTitleColor,
    bool isShowLastNextMonth,
    bool isShowSolarCalendar,
    Holiday holiday,
    Lunar lunar,
    Solar solar,
  }) : super(
    normalBackColor: normalBackColor, 
    selectedBackColor: selectedBackColor,
    normalTitleColor: normalTitleColor,
    selectedTitleColor: selectedTitleColor,
    otherMonthTitleColor: otherMonthTitleColor,
    isShowLastNextMonth: isShowLastNextMonth
  ) {
    this.holiday = holiday;
    this.lunar = lunar;
    this.solar = solar;
  }

  @override
  String get holidayName {
    if (holiday == null) {
      return '';
    }
    if (holiday.work) {
      return '班';
    }
    return '休';
  }
  
  @override
  String get lunarName {
    if (lunar == null) {
      return '';
    }
    if (lunar.day == '初一') {
      return lunar.month + lunar.day;
    }
    return lunar.day;
  }

  @override
  String get solarName {
    if (solar == null) {
      return '';
    }
    return solar.name;
  }
}

abstract class FxCalendarBuildModelDelegate {
  // 回调当前选中的时间
  void currentLayoutDateTime(DateTime dateTime);
  // 构建每天的数据模型
  FxItemModel buildCalendarModel(bool isLastNextMonth, int year, int month, int day);
}

const int FxCalendarShowMaxCount = 42;
// 计算日期
class FxCalendarManager {
  FxCalendarManager({
    this.initYear, 
    this.initMonth,
    this.fxDelegate
  }) {
    selectedDateTime = DateTime(initYear, initMonth);
  }
  // 初始化的年份，不可更改
  final int initYear;
  // 初始化的月份，不可更改
  final int initMonth;
  // 构建model对象方法
  final FxCalendarBuildModelDelegate fxDelegate;

  // 增加的月数，用户点击上下月
  int addMonthCount = 0;
  // 选中的日期
  DateTime selectedDateTime;
  DateTime lastDateTime;
  DateTime nextDateTime;
  // 显示选中月的全部数据
  List<FxItemModel> selectedMonthList = [];
  List<FxItemModel> lastMonthList = [];
  List<FxItemModel> nextMonthList = [];

  void updateDate(int count) {
    if (0 == count) {
      selectedDateTime = DateTime(initYear, initMonth);
      fxDelegate.currentLayoutDateTime(selectedDateTime);
      _loadDateTimeData(0);
      _loadDateTimeData(-1);
      _loadDateTimeData(1);
    }else if (1 == count) {
      // 下一个月
      addMonthCount += 1;
      lastDateTime = selectedDateTime;
      lastMonthList = selectedMonthList;
      selectedDateTime = nextDateTime;
      selectedMonthList = nextMonthList;
      nextMonthList = [];
      nextDateTime = null;
      fxDelegate.currentLayoutDateTime(selectedDateTime);
      _loadDateTimeData(1);
    }else if (-1 == count) {
      // 上一个月
      addMonthCount -= 1;
      nextDateTime = selectedDateTime;
      nextMonthList = selectedMonthList;
      selectedDateTime = lastDateTime;
      selectedMonthList = lastMonthList;
      lastMonthList = [];
      lastDateTime = null;
      fxDelegate.currentLayoutDateTime(selectedDateTime);
      _loadDateTimeData(-1);
    }
  }
  /// 根据选中的月份，更新数据
  /// [off] 月份偏移量。0：表示选中的月份；-1：表示上个月；1：表示下个月
  void _loadDateTimeData(int off) {
    // 1: 获取年月1日
    int nm1 = initMonth + addMonthCount + off;
    int nMonth = nm1 % 12;
    nMonth = nMonth == 0 ? 12 : nMonth;
    int nyCount = 0;  // 偏移量-年份
    if (nm1 > 0) {
      nyCount = nm1 ~/ 12;
      nyCount = (nm1 % 12) == 0 ? (nyCount - 1) : nyCount;
    }else if (nm1 == 0) {
      nyCount = -1;
    }else {
      nyCount = (nm1 ~/ 12) - 1;
    }
    int nYear = initYear + nyCount;
    DateTime nDateTime = DateTime(nYear, nMonth);
    // 3: 创建数据模型
    _loadDateTimeModel(nDateTime, off);
  }
  /// 处理数据模型
  /// [dateTime] 当前选中的日期时间
  /// [off] 偏移 0：表示当前时间；-1：上个月时间；1：下个月时间
  void _loadDateTimeModel(DateTime dateTime, int off) {
    List<FxItemModel> buildList = [];
    // 1: 处理上个月的部分数据
    int needCount = dateTime.weekday;
    needCount = needCount == 7 ? 0 : needCount;
    DateTime nLastDateTime = dateTime.subtract(Duration(days: 1));
    int nDayLast = nLastDateTime.day;
    for (int i = needCount - 1; i >= 0; i--) {
      FxItemModel model = fxDelegate.buildCalendarModel(true, nLastDateTime.year, nLastDateTime.month, nDayLast - i);
      buildList.add(model);
    }
    // 2: 处理选中月的全部数据 + 下个月的部分数据
    int nextMonthCount = 0;
    for (int i = 0; i < FxCalendarShowMaxCount - needCount; i++) {
      if (i < 28) {
        // 最少有28天
        FxItemModel model = fxDelegate.buildCalendarModel(false, dateTime.year, dateTime.month, i+1);
        buildList.add(model);
      }else {
        // 可能会到下一月
        DateTime mybeNextDateTime = dateTime.add(Duration(days: i));
        bool isNextMonth = dateTime.month != mybeNextDateTime.month;
        nextMonthCount += isNextMonth ? 1 : 0;
        FxItemModel model = fxDelegate.buildCalendarModel(isNextMonth, mybeNextDateTime.year, mybeNextDateTime.month, mybeNextDateTime.day);
        buildList.add(model);
      }
    }
    if (nextMonthCount >= 7) {
      buildList = buildList.sublist(0, buildList.length - 7);
    }
    if (off == 0) {
        selectedDateTime = dateTime;
        selectedMonthList = buildList;
      }else if (off == -1) {
        lastDateTime = dateTime;
        lastMonthList = buildList;
      }else {
        nextDateTime = dateTime;
        nextMonthList = buildList;
      }
  }
}